/* inc.c */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>

#include "main.h"
#include "inc.h"
#include "mainwindow.h"
#include "folderview.h"
#include "prefs_common.h"
#include "prefs_account.h"
#include "account.h"
#include "procmsg.h"
#include "socket.h"
#include "pop.h"
#include "recv.h"
#include "mbox.h"
#include "utils.h"
#include "gtkutils.h"
#include "statusbar.h"
#include "manage_window.h"
#include "progressdialog.h"
#include "inputdialog.h"
#include "alertpanel.h"
#include "filter.h"
#include "automaton.h"

#define MSGBUFSIZE	8192
#define TMPMSG 		".tmpmsg"

static void pop3_automaton_create	(PrefsAccount	*ac_prefs,
					 MainWindow	*mainwin);
static void pop3_automaton_destroy	(Automaton	*atm);
static gint pop3_terminate		(gint		 source,
					 Automaton	*atm);
#if USE_THREADS
static gint connection_check_cb		(Automaton	*atm);
#endif

static void inc_account_mail		(PrefsAccount	*ac_prefs,
					 gpointer	 data);

static ProgressDialog *inc_progress_create
					(gpointer	 data);
static void inc_cancel			(GtkWidget	*widget,
					 gpointer	 data);

static gint inc_spool			(void);
static gint get_spool			(const gchar	*dest,
					 const gchar	*mbox);

void inc_mail(MainWindow *mainwin)
{
	if (prefs_common.use_extinc && prefs_common.extinc_path) {
		gint pid;

		/* external incorporating program */
		if ((pid = fork()) < 0) {
			perror("fork");
			return;
		}

		if (pid == 0) {
			execlp(prefs_common.extinc_path,
			       g_basename(prefs_common.extinc_path),
			       NULL);

			/* this will be called when failed */
			perror("exec");
			_exit(1);
		}

		/* wait until child process is terminated */
		waitpid(pid, NULL, 0);

		if (prefs_common.inc_local) inc_spool();
	} else {
		if (prefs_common.inc_local) inc_spool();

		inc_account_mail(cur_account, mainwin);
	}

	folderview_unselect(mainwin->folderview);
	folderview_select(mainwin->folderview,
			  cur_account->inbox ? cur_account->inbox : INBOX_DIR);
}

void inc_all_account_mail(MainWindow *mainwin)
{
	if (prefs_common.inc_local) inc_spool();

	account_foreach(inc_account_mail, mainwin);

	folderview_unselect(mainwin->folderview);
	folderview_select(mainwin->folderview,
			  cur_account->inbox ? cur_account->inbox : INBOX_DIR);
}


static void pop3_automaton_create(PrefsAccount *ac_prefs, MainWindow *mainwin)
{
	Automaton *atm;
	Pop3State *state;
	SockInfo *sockinfo;
	gint i;
	gchar *server;
	gushort port;
	gchar *buf;

	atm = automaton_create(N_POP3_PHASE);

	state = g_new0(Pop3State, 1);
	atm->data = state;
	state->ac_prefs = ac_prefs;
	state->folder_table = g_hash_table_new(g_str_hash, g_str_equal);
	state->mainwin = mainwin;
	state->user = g_strdup(ac_prefs->userid);

	if (ac_prefs->passwd)
		state->pass = g_strdup(ac_prefs->passwd);
	else if (ac_prefs->tmp_pass)
		state->pass = g_strdup(ac_prefs->tmp_pass);
	else {
		gchar *pass;
		gchar *message;

		message = g_strdup_printf(_("Input password for %s:"),
					  state->user);

		pass = input_dialog_with_invisible(_("Input password"),
						   message, NULL);
		manage_window_focus_in(mainwin->window, NULL, NULL);
		if (pass) {
			ac_prefs->tmp_pass = g_strdup(pass);
			state->pass = pass;
		} else {
			pop3_terminate(-1, atm);
			return;
		}

		g_free(message);
	}

	state->progress = inc_progress_create(atm);

	atm->state[POP3_GREETING_RECV].handler =
		(AtmHandler)pop3_greeting_recv;
	atm->state[POP3_GETAUTH_USER_SEND].handler =
		(AtmHandler)pop3_getauth_user_send;
	atm->state[POP3_GETAUTH_USER_RECV].handler =
		(AtmHandler)pop3_getauth_user_recv;
	atm->state[POP3_GETAUTH_PASS_SEND].handler =
		(AtmHandler)pop3_getauth_pass_send;
	atm->state[POP3_GETAUTH_PASS_RECV].handler =
		(AtmHandler)pop3_getauth_pass_recv;
	atm->state[POP3_GETAUTH_APOP_SEND].handler =
		(AtmHandler)pop3_getauth_apop_send;
	atm->state[POP3_GETAUTH_APOP_RECV].handler =
		(AtmHandler)pop3_getauth_apop_recv;
	atm->state[POP3_GETRANGE_STAT_SEND].handler =
		(AtmHandler)pop3_getrange_stat_send;
	atm->state[POP3_GETRANGE_STAT_RECV].handler =
		(AtmHandler)pop3_getrange_stat_recv;
	atm->state[POP3_GETRANGE_LAST_SEND].handler =
		(AtmHandler)pop3_getrange_last_send;
	atm->state[POP3_GETRANGE_LAST_RECV].handler =
		(AtmHandler)pop3_getrange_last_recv;
	atm->state[POP3_RETR_SEND].handler = (AtmHandler)pop3_retr_send;
	atm->state[POP3_RETR_RECV].handler = (AtmHandler)pop3_retr_recv;
	atm->state[POP3_DELETE_SEND].handler = (AtmHandler)pop3_delete_send;
	atm->state[POP3_DELETE_RECV].handler = (AtmHandler)pop3_delete_recv;
	atm->state[POP3_LOGOUT_SEND].handler = (AtmHandler)pop3_logout_send;
	atm->state[POP3_LOGOUT_RECV].handler = (AtmHandler)pop3_logout_recv;

	atm->state[POP3_GREETING_RECV].condition = GDK_INPUT_READ;
	for (i = POP3_GETAUTH_USER_SEND; i < N_POP3_PHASE; ) {
		atm->state[i++].condition = GDK_INPUT_WRITE;
		atm->state[i++].condition = GDK_INPUT_READ;
	}

	atm->terminate = (AtmHandler)pop3_terminate;

	atm->num = POP3_GREETING_RECV;

	server = ac_prefs->recv_server;
	port = ac_prefs->set_popport ? ac_prefs->popport : 110;

	buf = g_strdup_printf(_("Connecting to POP3 server: %s ..."), server);
	log_message("%s\n", buf);
	progress_dialog_set_label(state->progress, buf);
	g_free(buf);
	GTK_EVENTS_FLUSH();

#if USE_THREADS
	if ((sockinfo = sock_connect_with_thread(server, port)) == NULL) {
#else
	if ((sockinfo = sock_connect_nb(server, port)) == NULL) {
#endif
		log_warning(_("Can't connect to POP3 server: %s:%d\n"),
			    server, port);
		manage_window_focus_in(state->progress->window, NULL, NULL);
		alertpanel_error(_("Can't connect to POP3 server: %s:%d\n"),
				 server, port);
		manage_window_focus_out(state->progress->window, NULL, NULL);
		pop3_terminate(-1, atm);
		return;
	}

	state->sockinfo = sockinfo;
	sockinfo->data = atm;

#if USE_THREADS
	atm->timeout_tag = gtk_timeout_add
		(TIMEOUT_ITV, (GtkFunction)connection_check_cb, atm);
#else
	atm->tag = gdk_input_add(sockinfo->sock,
				 atm->state[atm->num].condition,
				 automaton_input_cb, atm);
#endif

	gtk_main();

#if USE_THREADS
	//pthread_join(sockinfo->connect_thr, NULL);
#endif
	sock_sockinfo_free(sockinfo);
}

static void pop3_automaton_destroy(Automaton *atm)
{
	Pop3State *state = atm->data;

	g_hash_table_destroy(state->folder_table);
	g_free(state->greeting);
	g_free(state->user);
	g_free(state->pass);
	g_free(state->prev_folder);
	g_free(state);
	automaton_destroy(atm);
}

static gint pop3_terminate(gint source, Automaton *atm)
{
	Pop3State *state = atm->data;

	if (atm->tag > 0) gdk_input_remove(atm->tag);
	if (atm->timeout_tag > 0) gtk_timeout_remove(atm->timeout_tag);
	if (source > 0) sock_close(source);

	if (state->error_val == PS_AUTHFAIL) {
		manage_window_focus_in(state->progress->window, NULL, NULL);
		alertpanel_error(_("Authorization for %s failed"),
				 state->user);
	}

	statusbar_pop_all();
	progress_dialog_destroy(state->progress);
	manage_window_focus_in(state->mainwin->window, NULL, NULL);

	g_hash_table_foreach(state->folder_table,
			     folderview_scan_foreach_func,
			     state->mainwin->folderview);

	if (state->error_val == PS_AUTHFAIL && state->ac_prefs->tmp_pass) {
		g_free(state->ac_prefs->tmp_pass);
		state->ac_prefs->tmp_pass = NULL;
	}

	pop3_automaton_destroy(atm);

	if (gtk_main_level() > 1)
		gtk_main_quit();

	return 0;
}

#if USE_THREADS
static gint connection_check_cb(Automaton *atm)
{
	Pop3State *state = atm->data;
	SockInfo *sockinfo = state->sockinfo;

	//g_print("connection check\n");

	if (sockinfo->state == CONN_LOOKUPFAILED ||
	    sockinfo->state == CONN_FAILED) {
		atm->timeout_tag = 0;
		pop3_terminate(sockinfo->sock, atm);
		log_warning(_("Can't connect to POP3 server: %s:%d\n"),
			    sockinfo->hostname, sockinfo->port);
		manage_window_focus_in(state->progress->window, NULL, NULL);
		alertpanel_error(_("Can't connect to POP3 server: %s:%d\n"),
				 sockinfo->hostname, sockinfo->port);
		manage_window_focus_out(state->progress->window, NULL, NULL);
		return FALSE;
	} else if (sockinfo->state == CONN_ESTABLISHED) {
		atm->timeout_tag = 0;
		atm->tag = gdk_input_add(sockinfo->sock,
					 atm->state[atm->num].condition,
					 automaton_input_cb, atm);
		return FALSE;
	} else {
		return TRUE;
	}
}
#endif

void inc_progress_update(Pop3State *state, Pop3Phase phase)
{
	gchar buf[MSGBUFSIZE];

	switch (phase) {
	case POP3_GREETING_RECV:
		break;
	case POP3_GETAUTH_USER_SEND:
	case POP3_GETAUTH_USER_RECV:
	case POP3_GETAUTH_PASS_SEND:
	case POP3_GETAUTH_PASS_RECV:
	case POP3_GETAUTH_APOP_SEND:
	case POP3_GETAUTH_APOP_RECV:
		progress_dialog_set_label(state->progress, _("Authorizing"));
		break;
	case POP3_GETRANGE_STAT_SEND:
	case POP3_GETRANGE_STAT_RECV:
	case POP3_GETRANGE_LAST_SEND:
	case POP3_GETRANGE_LAST_RECV:
		progress_dialog_set_label(state->progress,
					  _("Getting number of new messages"));
		break;
	case POP3_RETR_SEND:
	case POP3_RETR_RECV:
		g_snprintf(buf, sizeof(buf),
			   _("Retrieving message (%d / %d)"),
			   state->cur_msg, state->count);
		progress_dialog_set_label(state->progress, buf);
		progress_dialog_set_percentage
			(state->progress,
			 (gfloat)state->cur_msg / state->count);
		break;
	case POP3_DELETE_SEND:
	case POP3_DELETE_RECV:
		progress_dialog_set_label(state->progress,
					  _("Deleting message"));
		break;
	case POP3_LOGOUT_SEND:
	case POP3_LOGOUT_RECV:
		progress_dialog_set_label(state->progress, _("Quitting"));
		break;
	}
}

gint inc_drop_message(const gchar *file, Pop3State *state)
{
	gchar *inbox;
	gchar *dropfolder;
	gchar *path;
	gint val;
	PrefsAccount *ac_prefs = state->ac_prefs;

	if (ac_prefs->inbox && *ac_prefs->inbox &&
	    is_dir_exist(ac_prefs->inbox))
		inbox = ac_prefs->inbox;
	else
		inbox = INBOX_DIR;

	if (ac_prefs->assort_on_recv) {
		dropfolder = filter_get_dest_folder
			(prefs_common.fltlist, file);
		if (!dropfolder || !is_dir_exist(dropfolder))
			dropfolder = inbox;
	} else
		dropfolder = inbox;

	if (strcmp2(state->prev_folder, dropfolder) != 0) {
		g_free(state->prev_folder);
		state->prev_folder = g_strdup(dropfolder);

		state->msgnum = procmsg_get_last_message_number(dropfolder);
		if (state->msgnum < 0) {
			g_warning(_("can't drop message into %s\n"),
				  dropfolder);
			return -1;
		}
	}

	state->msgnum++;

	path = g_strdup_printf("%s%c%d", dropfolder,G_DIR_SEPARATOR,
			       state->msgnum);
	if (is_file_exist(path)) {
		g_warning(_("%s exists\n"), path);
		g_free(path);
		return -1;
	}

	if (rename(file, path) < 0) {
		FILE_OP_ERROR(file, "rename");
		g_warning("can't move tmpmsg to %s\n", path);
		unlink(file);
		g_free(path);
		return -1;
	}
	g_free(path);

	if ((val = GPOINTER_TO_INT(g_hash_table_lookup(state->folder_table,
						       dropfolder))))
		g_hash_table_insert(state->folder_table, dropfolder,
				    GINT_TO_POINTER(val + 1));
	else
		g_hash_table_insert(state->folder_table, g_strdup(dropfolder),
				    GINT_TO_POINTER(1));

	return 0;
}

static void inc_account_mail(PrefsAccount *ac_prefs, gpointer data)
{
	MainWindow *mainwin = (MainWindow *)data;

	if (!ac_prefs) return;
	if ((ac_prefs->protocol == A_POP3 ||
	     ac_prefs->protocol == A_APOP) &&
	     ac_prefs->recv_server) {
		debug_print(_("getting new messages of account %s...\n"),
			    ac_prefs->account_name);

		pop3_automaton_create(ac_prefs, mainwin);
	}
}

static ProgressDialog *inc_progress_create(gpointer data)
{
	ProgressDialog *progress;

	progress = progress_dialog_create();
	gtk_window_set_title(GTK_WINDOW(progress->window),
			     _("Retrieving new messages"));
	gtk_signal_connect(GTK_OBJECT(progress->cancel_btn), "clicked",
			   GTK_SIGNAL_FUNC(inc_cancel), data);
	gtk_signal_connect(GTK_OBJECT(progress->window), "delete_event",
			   GTK_SIGNAL_FUNC(gtk_true), NULL);
	manage_window_set_transient(GTK_WINDOW(progress->window));

	progress_dialog_set_value(progress, 0.0);

	gtk_widget_show_now(progress->window);

	return progress;
}

static void inc_cancel(GtkWidget *widget, gpointer data)
{
	Automaton *atm = (Automaton *)data;
	Pop3State *state = atm->data;
	SockInfo *sockinfo = state->sockinfo;

#if USE_THREADS
	if (sockinfo->state == CONN_READY ||
	    sockinfo->state == CONN_LOOKUPSUCCESS) {
		pthread_cancel(sockinfo->connect_thr);
		//pthread_kill(sockinfo->connect_thr, SIGINT);
		g_print("connection was cancelled.\n");
	}
#endif

	pop3_terminate(sockinfo->sock, atm);
}

static gint inc_spool(void)
{
	gchar *mbox, *logname;
	gint msgs;

	logname = g_get_user_name();
	mbox = g_strconcat(prefs_common.spool_path
			   ? prefs_common.spool_path : DEFAULT_SPOOL_PATH,
			   G_DIR_SEPARATOR_S, logname, NULL);
	msgs = get_spool(NULL, mbox);
	g_free(mbox);

	return msgs;
}

static gint get_spool(const gchar *dest, const gchar *mbox)
{
	gint msgs, size;
	gint lockfd;
	gchar *tmp_mbox = "/tmp/tmpmbox";

	g_return_val_if_fail(mbox != NULL, -1);
	CHDIR_RETURN_VAL_IF_FAIL(maildir, -1);

	if (!dest) dest = INBOX_DIR;

	if (!is_file_exist(mbox) || (size = get_file_size(mbox)) == 0) {
		debug_print(_("no messages in local mailbox.\n"));
		return 0;
	} else if (size < 0)
		return -1;

	if ((lockfd = lock_mbox(mbox, LOCK_FLOCK)) < 0)
		return -1;

	if (copy_mbox(mbox, tmp_mbox) < 0)
		return -1;

	debug_print(_("Getting new messages from %s into %s...\n"),
		    mbox, dest);
	if ((msgs = proc_mbox(dest, tmp_mbox)) < 0) {
		unlink(tmp_mbox);
		unlock_mbox(mbox, lockfd, LOCK_FLOCK);
		return -1;
	}

	unlink(tmp_mbox);
	empty_mbox(mbox);
	unlock_mbox(mbox, lockfd, LOCK_FLOCK);

	return msgs;
}
