/* messageview.c */

#include <glib.h>
#include <gtk/gtkvpaned.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkeditable.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "main.h"
#include "messageview.h"
#include "headerview.h"
#include "textview.h"
#include "mimeview.h"
#include "procmime.h"
#include "procheader.h"
#include "utils.h"

static void messageview_change_view_type(MessageView	*messageview,
					 MessageType	 type);

MessageView *messageview_create(void)
{
	MessageView *messageview;
	GtkWidget *paned;
	HeaderView *headerview;
	TextView *textview;
	MimeView *mimeview;

	debug_print(_("Creating message view...\n"));
	messageview = g_new0(MessageView, 1);

	messageview->type = MVIEW_TEXT;

	headerview = headerview_create();

	textview = textview_create();
	textview->messageview = messageview;

	mimeview = mimeview_create();
	mimeview->textview = textview;
	mimeview->messageview = messageview;

	paned = gtk_vpaned_new();
	gtk_paned_add1(GTK_PANED(paned), *(GtkWidget **)headerview);
	gtk_paned_add2(GTK_PANED(paned), *(GtkWidget **)textview);

	/* to remove without destroyed */
	gtk_widget_ref(*(GtkWidget **)textview);
	gtk_widget_ref(*(GtkWidget **)mimeview);

	messageview->paned      = paned;
	messageview->headerview = headerview;
	messageview->textview   = textview;
	messageview->mimeview   = mimeview;

	return messageview;
}

void messageview_init(MessageView *messageview)
{
	headerview_init(messageview->headerview);
	textview_init(messageview->textview);
	mimeview_init(messageview->mimeview);
	messageview_set_font();
}

void messageview_show(MessageView *messageview, const gchar *file)
{
	FILE *fp;
	MimeInfo *mimeinfo;
	MsgInfo *msginfo;

	g_return_if_fail(file != NULL);

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return;
	}
	mimeinfo = procmime_scan_mime_header(fp);
	fclose(fp);
	msginfo = procheader_parse(file);
	headerview_show(messageview->headerview, msginfo);
	procmsg_msginfo_free(msginfo);

	if (mimeinfo && mimeinfo->mime_type == MIME_MULTIPART) {
		messageview_change_view_type(messageview, MVIEW_MIME);
		mimeview_show_message(messageview->mimeview, mimeinfo, file);
	} else {
		messageview_change_view_type(messageview, MVIEW_TEXT);
		textview_show_message(messageview->textview, mimeinfo, file);
		procmime_mimeinfo_free(mimeinfo);
		gtk_widget_grab_focus
			(messageview->mainwin->summaryview->ctree);
	}
}

static void messageview_change_view_type(MessageView *messageview,
					 MessageType type)
{
	TextView *textview = messageview->textview;
	MimeView *mimeview = messageview->mimeview;

	if (messageview->type == type) return;

	if (messageview->type == MVIEW_TEXT) {
		gtk_container_remove(GTK_CONTAINER(*(GtkWidget **)messageview),
				     *(GtkWidget **)textview);
		gtk_paned_add2(GTK_PANED(messageview->paned),
			       *(GtkWidget **)mimeview);
		gtk_container_add(GTK_CONTAINER(mimeview->vbox),
				  *(GtkWidget **)textview);

		messageview->type = MVIEW_MIME;
	} else if (messageview->type == MVIEW_MIME) {
		gtk_container_remove(GTK_CONTAINER(*(GtkWidget **)messageview),
				     *(GtkWidget **)mimeview);
		gtk_container_remove(GTK_CONTAINER(mimeview->vbox),
				     *(GtkWidget **)textview);
		gtk_paned_add2(GTK_PANED(messageview->paned),
			       *(GtkWidget **)textview);

		messageview->type = MVIEW_TEXT;
	}
}

void messageview_clear(MessageView *messageview)
{
	messageview_change_view_type(messageview, MVIEW_TEXT);
	headerview_clear(messageview->headerview);
	textview_clear(messageview->textview);
}

void messageview_quote_color_set(void)
{
}

void messageview_set_font(void)
{
	textview_set_font();
}

void messageview_copy_clipboard(MessageView *messageview)
{
	if (messageview->type == MVIEW_TEXT)
		gtk_editable_copy_clipboard
			(GTK_EDITABLE(messageview->textview->text));
}

void messageview_select_all(MessageView *messageview)
{
	if (messageview->type == MVIEW_TEXT)
		gtk_editable_select_region
			(GTK_EDITABLE(messageview->textview->text), 0, -1);
}

GtkWidget *messageview_get_text_widget(MessageView *messageview)
{
	return messageview->textview->text;
}
