/* mimeview.c */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkctree.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkvpaned.h>
#include <gtk/gtktext.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkmenu.h>
#include <stdio.h>

#include "intl.h"
#include "main.h"
#include "mimeview.h"
#include "procmime.h"
#include "summaryview.h"
#include "menu.h"
#include "filesel.h"
#include "utils.h"
#include "gtkutils.h"

typedef enum
{
	COL_MIMETYPE = 0,
	COL_SIZE     = 1,
	COL_NAME     = 2
} MimeViewColumnPos;

#define N_MIMEVIEW_COLS	3

static void mimeview_scan_multipart_message	(MimeView	*mimeview,
						 MimeInfo	*mimeinfo,
						 GtkCTreeNode	*parent,
						 FILE		*fp);
static GtkCTreeNode *mimeview_append_part	(MimeView	*mimeview,
						 MimeInfo	*partinfo,
						 GtkCTreeNode	*parent);
static void mimeview_show_message_part		(MimeView	*mimeview,
						 MimeInfo	*partinfo);
static void mimeview_clear			(MimeView	*mimeview);

static void mimeview_selected		(GtkCTree	*ctree,
					 GtkCTreeNode	*node,
					 gint		 column,
					 MimeView	*mimeview);
static void mimeview_button_pressed	(GtkWidget	*widget,
					 GdkEventButton	*event,
					 MimeView	*mimeview);
static void mimeview_key_pressed	(GtkWidget	*widget,
					 GdkEventKey	*event,
					 MimeView	*mimeview);

static void mimeview_display_as_text	(MimeView	*mimeview);
static void mimeview_save_as		(MimeView	*mimeview);

static GtkItemFactoryEntry mimeview_popup_entries[] =
{
	{N_("/_Display as text"),	NULL, mimeview_display_as_text, 0, NULL},
	{N_("/_Save as..."),		NULL, mimeview_save_as,	0, NULL}
};

MimeView *mimeview_create(void)
{
	MimeView *mimeview;

	GtkWidget *paned;
	GtkWidget *scrolledwin;
	GtkWidget *ctree;
	GtkWidget *vbox;
	GtkWidget *popupmenu;
	GtkItemFactory *popupfactory;

	gchar *titles[] = {_("MIME Type"), _("Size"), _("Name")};
	gint n_entries;
	gint i;

	debug_print(_("Creating MIME view...\n"));
	mimeview = g_new0(MimeView, 1);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_ALWAYS);
	gtk_widget_set_usize(scrolledwin, -1, 80);

	ctree = gtk_ctree_new_with_titles(N_MIMEVIEW_COLS, 0, titles);
	gtk_clist_set_selection_mode(GTK_CLIST(ctree), GTK_SELECTION_BROWSE);
	gtk_ctree_set_line_style(GTK_CTREE(ctree), GTK_CTREE_LINES_NONE);
	gtk_clist_set_column_justification(GTK_CLIST(ctree), COL_SIZE,
					   GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_width(GTK_CLIST(ctree), COL_MIMETYPE, 180);
	gtk_clist_set_column_width(GTK_CLIST(ctree), COL_SIZE, 48);
	for (i = 0; i < N_MIMEVIEW_COLS; i++)
		GTK_WIDGET_UNSET_FLAGS(GTK_CLIST(ctree)->column[i].button,
				       GTK_CAN_FOCUS);
	gtk_container_add(GTK_CONTAINER(scrolledwin), ctree);

	gtk_signal_connect(GTK_OBJECT(ctree), "tree_select_row",
			   GTK_SIGNAL_FUNC(mimeview_selected), mimeview);
	gtk_signal_connect(GTK_OBJECT(ctree), "button_press_event",
			   GTK_SIGNAL_FUNC(mimeview_button_pressed), mimeview);
	gtk_signal_connect(GTK_OBJECT(ctree), "key_press_event",
			   GTK_SIGNAL_FUNC(mimeview_key_pressed), mimeview);

	vbox = gtk_vbox_new(FALSE, 0);

	paned = gtk_vpaned_new();
	gtk_paned_add1(GTK_PANED(paned), scrolledwin);
	gtk_paned_add2(GTK_PANED(paned), vbox);
	gtk_widget_show_all(paned);

	n_entries = sizeof(mimeview_popup_entries) /
		sizeof(mimeview_popup_entries[0]);
	popupmenu = menu_create_items(mimeview_popup_entries, n_entries,
				      "<MimeView>", &popupfactory, mimeview);

	mimeview->paned       = paned;
	mimeview->scrolledwin = scrolledwin;
	mimeview->ctree       = ctree;
	mimeview->vbox        = vbox;
	mimeview->popupmenu   = popupmenu;
	mimeview->popupfactory = popupfactory;

	return mimeview;
}

void mimeview_init(MimeView *mimeview)
{
}

void mimeview_show_message(MimeView *mimeview, MimeInfo *mimeinfo,
			   const gchar *file)
{
	GtkCTree *ctree = GTK_CTREE(mimeview->ctree);
	GtkCTreeNode *node;
	gchar buf[BUFFSIZE];
	FILE *fp;

	mimeview_clear(mimeview);

	g_return_if_fail(file != NULL);
	g_return_if_fail(mimeinfo != NULL);

	mimeview->mimeinfo = mimeinfo;

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return;
	}
	mimeview->file = g_strdup(file);

	/* skip headers */
	while (fgets(buf, sizeof(buf), fp) != NULL)
		if (buf[0] == '\r' || buf[0] == '\n') break;

	gtk_signal_handler_block_by_func(GTK_OBJECT(ctree), mimeview_selected,
					 mimeview);

	mimeview_scan_multipart_message(mimeview, mimeinfo, NULL, fp);
	fclose(fp);

	gtk_signal_handler_unblock_by_func(GTK_OBJECT(ctree),
					   mimeview_selected, mimeview);

	if ((node = gtk_ctree_node_nth(ctree, 0))) {
		gtk_ctree_select(ctree, node);
		gtk_widget_grab_focus(mimeview->ctree);
	}
}

static void mimeview_scan_multipart_message(MimeView *mimeview,
					    MimeInfo *mimeinfo,
					    GtkCTreeNode *parent, FILE *fp)
{
	gchar *p;
	gchar *boundary;
	gint boundary_len;
	gchar buf[BUFFSIZE];
	glong fpos, prev_fpos;
	gint npart;
	static gint level;

	g_return_if_fail(mimeinfo != NULL);
	g_return_if_fail(mimeinfo->mime_type == MIME_MULTIPART);
	g_return_if_fail(mimeinfo->boundary != NULL);
	g_return_if_fail(mimeinfo->sub == NULL);
	g_return_if_fail(fp != NULL);

	if (!parent) level = 0;

	boundary = mimeinfo->boundary;
	boundary_len = strlen(boundary);

	/* look for first boundary */
	while ((p = fgets(buf, sizeof(buf), fp)) != NULL)
		if (IS_BOUNDARY(buf, boundary, boundary_len)) break;
	if (p == NULL) return;

	if ((fpos = ftell(fp)) < 0) {
		perror("ftell");
		return;
	}

	for (npart = 0;; npart++) {
		MimeInfo *partinfo;
		gboolean eom = FALSE;
		GtkCTreeNode *node = NULL;

		prev_fpos = fpos;

		partinfo = procmime_scan_mime_header(fp);
		if (!partinfo) break;
		partinfo->parent = mimeinfo;
		partinfo->fpos = fpos;

#if 0
		if (npart == 0 && mimeinfo->parent &&
		    partinfo->mime_type == MIME_TEXT)
			mimeinfo->sub = partinfo;
		else
#endif
			node = mimeview_append_part(mimeview, partinfo,
						    parent);

		if (partinfo->mime_type == MIME_MULTIPART) {
			level++;
			if (level < 8)
				mimeview_scan_multipart_message
					(mimeview, partinfo, node, fp);
			level--;
		} else if (partinfo->mime_type == MIME_MESSAGE_RFC822) {
			MimeInfo *sub;

			if ((fpos = ftell(fp)) < 0) {
				perror("ftell");
				procmime_mimeinfo_free(partinfo);
				break;
			}
			partinfo->sub = sub = procmime_scan_mime_header(fp);
			if (!sub) break;
			sub->fpos = fpos;

			if (sub->mime_type == MIME_MULTIPART) {
				level++;
				if (level < 8)
					mimeview_scan_multipart_message
						(mimeview, sub, node, fp);
				level--;
			}
		}

		/* look for next boundary */
		buf[0] = '\0';
		while ((p = fgets(buf, sizeof(buf), fp)) != NULL) {
			if (IS_BOUNDARY(buf, boundary, boundary_len)) {
				if (buf[2 + boundary_len]     == '-' &&
				    buf[2 + boundary_len + 1] == '-')
					eom = TRUE;
				break;
			}
		}
		if ((fpos = ftell(fp)) < 0) {
			perror("ftell");
			procmime_mimeinfo_free(partinfo);
			break;
		}

		partinfo->size = fpos - prev_fpos - strlen(buf);

		if (node)
			gtk_ctree_node_set_text
				(GTK_CTREE(mimeview->ctree), node, COL_SIZE,
				 to_human_readable(partinfo->size));

		if (eom) break;
	}
}

static GtkCTreeNode *mimeview_append_part(MimeView *mimeview,
					  MimeInfo *partinfo,
					  GtkCTreeNode *parent)
{
	GtkCTree *ctree = GTK_CTREE(mimeview->ctree);
	GtkCTreeNode *node;
	gchar *str[N_MIMEVIEW_COLS];

	str[COL_MIMETYPE] =
		partinfo->content_type ? partinfo->content_type : "";
	str[COL_SIZE] = "";
	if (partinfo->name)
		str[COL_NAME] = partinfo->name;
	else if (partinfo->filename)
		str[COL_NAME] = partinfo->filename;
	else
		str[COL_NAME] = "";

	node = gtk_ctree_insert_node(ctree, parent, NULL, str, 0,
				     NULL, NULL, NULL, NULL,
				     FALSE, TRUE);
	gtk_ctree_node_set_row_data(ctree, node, partinfo);

	return node;
}

static void mimeview_show_message_part(MimeView *mimeview, MimeInfo *partinfo)
{
	FILE *fp;

	if (!partinfo) return;
	if (!mimeview->file) return;

	if ((fp = fopen(mimeview->file, "r")) == NULL) {
		FILE_OP_ERROR(mimeview->file, "fopen");
		return;
	}

	if (fseek(fp, partinfo->fpos, SEEK_SET) < 0) {
		FILE_OP_ERROR(mimeview->file, "fseek");
		fclose(fp);
		return;
	}

	textview_show_part(mimeview->textview, partinfo, fp);

	fclose(fp);
}

static void mimeview_clear(MimeView *mimeview)
{
	GtkCList *clist = GTK_CLIST(mimeview->ctree);
	GtkCTree *ctree = GTK_CTREE(mimeview->ctree);
	GtkCTreeNode *node;
	MimeInfo *partinfo;

	for (node = GTK_CTREE_NODE(clist->row_list); node != NULL;
	     node = GTK_CTREE_NODE_NEXT(node)) {
		partinfo = gtk_ctree_node_get_row_data(ctree, node);
		procmime_mimeinfo_free(partinfo);
	}

	procmime_mimeinfo_free(mimeview->mimeinfo);

	gtk_clist_clear(clist);
	textview_clear(mimeview->textview);

	mimeview->opened = NULL;

	g_free(mimeview->file);
	mimeview->file = NULL;
}

static void mimeview_selected(GtkCTree *ctree, GtkCTreeNode *node, gint column,
			      MimeView *mimeview)
{
	MimeInfo *partinfo;

	if (mimeview->opened == node) return;
	mimeview->opened = node;

	partinfo = gtk_ctree_node_get_row_data(ctree, node);

	if (partinfo && (partinfo->mime_type == MIME_TEXT ||
			 partinfo->mime_type == MIME_MESSAGE_RFC822 ||
			 partinfo->mime_type == MIME_MULTIPART))
		mimeview_show_message_part(mimeview, partinfo);
	else
		textview_show_mime_part(mimeview->textview, partinfo);
}

static void mimeview_button_pressed(GtkWidget *widget, GdkEventButton *event,
				    MimeView *mimeview)
{
	GtkCList *clist = GTK_CLIST(widget);
	MimeInfo *partinfo;
	gint row, column;

	if (!event) return;

	if (event->button == 1) return;

	if (event->button == 2 || event->button == 3) {
		if (!gtk_clist_get_selection_info(clist, event->x, event->y,
						  &row, &column))
			return;
		gtk_clist_unselect_all(clist);
		gtk_clist_select_row(clist, row, column);
		gtkut_clist_set_focus_row(clist, row);
	}

	if (event->button != 3) return;

	partinfo = gtk_ctree_node_get_row_data
		(GTK_CTREE(mimeview->ctree), mimeview->opened);
	if (partinfo && (partinfo->mime_type == MIME_TEXT ||
			 partinfo->mime_type == MIME_MESSAGE_RFC822 ||
			 partinfo->mime_type == MIME_MULTIPART))
		menu_set_sensitive(mimeview->popupfactory,
				   "/Display as text", FALSE);
	else
		menu_set_sensitive(mimeview->popupfactory,
				   "/Display as text", TRUE);

	gtk_menu_popup(GTK_MENU(mimeview->popupmenu), NULL, NULL, NULL, NULL,
		       event->button, event->time);
}

#define BREAK_ON_MODIFIER_KEY() \
	if ((event->state & (GDK_MOD1_MASK|GDK_CONTROL_MASK)) != 0) break

static void mimeview_key_pressed(GtkWidget *widget, GdkEventKey *event,
				 MimeView *mimeview)
{
	SummaryView *summaryview = mimeview->messageview->mainwin->summaryview;
	GtkCTree *ctree = GTK_CTREE(mimeview->ctree);
	GtkCList *clist = GTK_CLIST(mimeview->ctree);
	GtkText *text;
	GtkCTreeNode *node;
	gfloat upper;

	if (!event) return;

	text = GTK_TEXT(mimeview->textview->text);

	switch (event->keyval) {
	case GDK_space:
		upper = text->vadj->upper - text->vadj->page_size;
		if (text->vadj->value < upper) {
			text->vadj->value += text->vadj->page_increment;
			text->vadj->value = MIN(text->vadj->value, upper);
			gtk_signal_emit_by_name(GTK_OBJECT(text->vadj),
						"value_changed");
			break;
		}

		node = GTK_CTREE_NODE_NEXT(mimeview->opened);
		if (node) {
			gtk_ctree_node_moveto(ctree, node, -1, 0.5, 0);
			gtk_clist_unselect_all(clist);
			gtk_ctree_select(ctree, node);
			gtkut_ctree_set_focus_row(ctree, node);
		} else
			summary_pass_key_press_event(summaryview, event);
		break;
	case GDK_n:
	case GDK_N:
		BREAK_ON_MODIFIER_KEY();
		if (!GTK_CTREE_NODE_NEXT(mimeview->opened)) {
			summary_pass_key_press_event(summaryview, event);
			break;
		}

		gtk_signal_emit_by_name(GTK_OBJECT(ctree), "scroll_vertical",
					GTK_SCROLL_STEP_FORWARD, 0.0);
		gtk_ctree_node_moveto(ctree, mimeview->opened, -1, 0.5, 0);
		break;
	case GDK_p:
	case GDK_P:
		BREAK_ON_MODIFIER_KEY();
		if (!GTK_CTREE_NODE_PREV(mimeview->opened)) {
			summary_pass_key_press_event(summaryview, event);
			break;
		}

		gtk_signal_emit_by_name(GTK_OBJECT(ctree), "scroll_vertical",
					GTK_SCROLL_STEP_BACKWARD, 0.0);
		gtk_ctree_node_moveto(ctree, mimeview->opened, -1, 0.5, 0);
		break;
	case GDK_y:
		BREAK_ON_MODIFIER_KEY();
		mimeview_save_as(mimeview);
		break;
	case GDK_t:
		BREAK_ON_MODIFIER_KEY();
		mimeview_display_as_text(mimeview);
		break;
	default:
		summary_pass_key_press_event(summaryview, event);
	}
}

static void mimeview_display_as_text(MimeView *mimeview)
{
	MimeInfo *partinfo;

	if (!mimeview->opened) return;

	partinfo = gtk_ctree_node_get_row_data
		(GTK_CTREE(mimeview->ctree), mimeview->opened);
	mimeview_show_message_part(mimeview, partinfo);
}

static void mimeview_save_as(MimeView *mimeview)
{
	gchar *filename;
	FILE *fp, *outfp;
	MimeInfo *partinfo;
	gchar buf[BUFFSIZE];

	if (!mimeview->opened) return;
	if (!mimeview->file) return;

	partinfo = gtk_ctree_node_get_row_data
		(GTK_CTREE(mimeview->ctree), mimeview->opened);
	g_return_if_fail(partinfo != NULL);

	filename = filesel_select_file
		(_("Save as"),
		 partinfo->filename ? partinfo->filename : partinfo->name);
	if (!filename) return;

	debug_print(_("Saving part of multipart message to %s...\n"),
		    filename);

	if ((fp = fopen(mimeview->file, "r")) == NULL) {
		FILE_OP_ERROR(mimeview->file, "fopen");
		return;
	}
	if (fseek(fp, partinfo->fpos, SEEK_SET) < 0) {
		FILE_OP_ERROR(mimeview->file, "fseek");
		fclose(fp);
		return;
	}

	if ((outfp = fopen(filename, "w")) == NULL) {
		FILE_OP_ERROR(filename, "fopen");
		fclose(fp);
		return;
	}

	while (fgets(buf, sizeof(buf), fp) != NULL)
		if (buf[0] == '\r' || buf[0] == '\n') break;

	procmime_decode_content(outfp, fp, partinfo);

	fclose(outfp);
	fclose(fp);

	debug_print(_("done.\n"));
}
