/* send.c */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <stdio.h>
#include <string.h>

#include "intl.h"
#include "main.h"
#include "send.h"
#include "socket.h"
#include "smtp.h"
#include "account.h"
#include "compose.h"
#include "procmsg.h"
#include "procheader.h"
#include "utils.h"

static gint send_message_smtp	(GSList *to_list, const gchar *from,
				 const gchar *server, FILE *fp);
static gint send_smtp_open	(void);

#define SEND_EXIT_IF_ERROR(f, s) \
{ \
	if ((f) < 0) { \
		log_warning("Error occured while %s\n", s); \
		if (smtp_sock > 0) sock_close(smtp_sock); \
		return -1; \
	} \
}

#define SEND_EXIT_IF_NOTOK(f, s) \
{ \
	if ((f) != SM_OK) { \
		log_warning("Error occured while %s\n", s); \
		if (smtp_quit(smtp_sock) != SM_OK) \
			log_warning("Error occured while sending QUIT\n"); \
		sock_close(smtp_sock); \
		return -1; \
	} \
}

gint send_message(GSList *to_list, const gchar *file)
{
	FILE *fp;
	gint val;

	g_return_val_if_fail(to_list != NULL, -1);
	g_return_val_if_fail(file != NULL, -1);
	g_return_val_if_fail(cur_account != NULL, -1);

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return -1;
	}

	val = send_message_smtp(to_list, cur_account->address,
				cur_account->smtp_server, fp);

	fclose(fp);
	return val;
}

enum
{
	Q_SENDER     = 0,
	Q_SMTPSERVER = 1,
	Q_RECIPIENTS = 2
};

gint send_message_queue(const gchar *file)
{
	static HeaderEntry qentry[] = {{"S:",   NULL, FALSE},
				       {"SSV:", NULL, FALSE},
				       {"R:",   NULL, FALSE},
				       {NULL,   NULL, FALSE}};
	FILE *fp;
	gint val;
	gchar *from = NULL;
	gchar *server = NULL;
	GSList *to_list = NULL;
	gchar buf[BUFFSIZE];
	gint hnum;

	g_return_val_if_fail(file != NULL, -1);

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return -1;
	}

	while ((hnum = procheader_get_one_field(buf, sizeof(buf), fp, qentry))
	       != -1) {
		gchar *p = buf + strlen(qentry[hnum].name);

		switch (hnum) {
		case Q_SENDER:
			from = g_strdup(p);
			break;
		case Q_SMTPSERVER:
			server = g_strdup(p);
			break;
		case Q_RECIPIENTS:
			to_list = compose_to_list_append(to_list, p);
			break;
		default:
		}
	}

	if (!to_list || !from || !server) {
		g_warning(_("Queued message header is broken.\n"));
		val = -1;
	} else {
		val = send_message_smtp(to_list, from, server, fp);
		if (val == 0) {
			if (unlink(file) < 0)
				FILE_OP_ERROR(file, "unlink");
		}
	}

	compose_to_list_remove_all(to_list);
	g_free(from);
	g_free(server);
	fclose(fp);

	return val;
}

static gint send_message_smtp(GSList *to_list, const gchar *from,
			      const gchar *server, FILE *fp)
{
	gint smtp_sock;
	gchar buf[BUFFSIZE];
	GSList *cur;

	g_return_val_if_fail(to_list != NULL, -1);
	g_return_val_if_fail(from != NULL, -1);
	g_return_val_if_fail(server != NULL, -1);
	g_return_val_if_fail(fp != NULL, -1);

	log_message(_("Connecting to SMTP server: %s ...\n"), server);

	SEND_EXIT_IF_ERROR((smtp_sock = send_smtp_open()),
			   "connecting to server");
	SEND_EXIT_IF_NOTOK(smtp_from(smtp_sock, from), "sending MAIL FROM");
	for (cur = to_list; cur != NULL; cur = cur->next)
		SEND_EXIT_IF_NOTOK(smtp_rcpt(smtp_sock, (gchar *)cur->data),
				   "sending RCPT TO");
	SEND_EXIT_IF_NOTOK(smtp_data(smtp_sock), "sending DATA");

	/* send main part */
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		strretchomp(buf);

		/* escape when a dot appears on the top */
		if (buf[0] == '.')
			SEND_EXIT_IF_ERROR(sock_write(smtp_sock, ".", 1),
					   "sending data");

		SEND_EXIT_IF_ERROR(sock_puts(smtp_sock, buf), "sending data");
	}

	SEND_EXIT_IF_NOTOK(smtp_eom(smtp_sock), "terminating data");
	SEND_EXIT_IF_NOTOK(smtp_quit(smtp_sock), "sending QUIT");

	sock_close(smtp_sock);
	return 0;
}

static gint send_smtp_open(void)
{
	SockInfo *sockinfo;
	gint sock;
	const gchar *server;
	gushort port;
	gint val;

	g_return_val_if_fail(cur_account != NULL, -1);

	server = cur_account->smtp_server;
	port = cur_account->set_smtpport ? cur_account->smtpport : 25;

	if ((sockinfo = sock_connect(server, port)) == NULL) {
		log_warning(_("Can't connect to SMTP server: %s:%d\n"),
			    server, port);
		return -1;
	}
	sock = sockinfo->sock;
	sock_sockinfo_free(sockinfo);

	if (smtp_ok(sock) == SM_OK) {
		if (cur_account->set_domain &&
		    cur_account->domain && *cur_account->domain)
			val = smtp_helo(sock, cur_account->domain);
		else
			val = smtp_helo(sock, get_domain_name());

		if (val == SM_OK) return sock;
	}

	log_warning(_("Error occured while sending HELO\n"));
	sock_close(sock);

	return -1;
}
