/* textview.c */

#include <glib.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtktext.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "main.h"
#include "summaryview.h"
#include "procheader.h"
#include "prefs_common.h"
#include "utils.h"
#include "gtkutils.h"
#include "procmime.h"
#include "unmime.h"

typedef struct _RemoteURI	RemoteURI;

struct _RemoteURI
{
	gchar *uri;

	guint start;
	guint end;
};

static GdkColor quote_color = {
	(gulong)0,
	(gushort)0,
	(gushort)0,
	(gushort)0xbfff
};

static GdkColor uri_color = {
	(gulong)0,
	(gushort)0,
	(gushort)0x8fff,
	(gushort)0
};

static GdkColor emphasis_color = {
	(gulong)0,
	(gushort)0,
	(gushort)0,
	(gushort)0xcfff
};

static GdkFont *msgfont;
static GdkFont *normalfont;
static GdkFont *boldfont;

static void textview_write_line		(TextView	*textview,
					 const gchar	*str,
					 gboolean	 charset_is_jis);
static void textview_scan_header	(TextView	*textview,
					 FILE		*fp);

static void textview_key_pressed	(GtkWidget	*widget,
					 GdkEventKey	*event,
					 TextView	*textview);
static void textview_button_pressed	(GtkWidget	*widget,
					 GdkEventButton	*event,
					 TextView	*textview);

static void textview_uri_list_remove_all(GSList		*uri_list);
static void textview_open_uri		(const gchar	*uri);

TextView *textview_create(void)
{
	TextView *textview;
	GtkWidget *scrolledwin;
	GtkWidget *text;

	debug_print(_("Creating text view...\n"));
	textview = g_new0(TextView, 1);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

	text = gtk_text_new(gtk_scrolled_window_get_hadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)),
			    gtk_scrolled_window_get_vadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)));
	gtk_widget_set_usize(scrolledwin, prefs_common.mainview_width, -1);
	gtk_container_add(GTK_CONTAINER(scrolledwin), text);
	gtk_signal_connect(GTK_OBJECT(text), "key_press_event",
			   GTK_SIGNAL_FUNC(textview_key_pressed),
			   textview);
	gtk_signal_connect(GTK_OBJECT(text), "button_press_event",
			   GTK_SIGNAL_FUNC(textview_button_pressed),
			   textview);

	gtk_widget_show_all(scrolledwin);

	textview->scrolledwin = scrolledwin;
	textview->text        = text;
	textview->uri_list    = NULL;

	return textview;
}

void textview_init(TextView *textview)
{
	GdkColormap *colormap;
	GdkColor color[3];
	gboolean success[3];

	gtkut_widget_disable_theme_engine(textview->text);

	color[0] = quote_color;
	color[1] = uri_color;
	color[2] = emphasis_color;
	colormap = gdk_window_get_colormap(textview->text->window);
	gdk_colormap_alloc_colors(colormap, color, 3, FALSE, TRUE, success);
	if (success[0] == FALSE || success[1] == FALSE ||
	    success[2] == FALSE) {
		GtkStyle *style;

		g_warning(_("TextView: color allocation failed\n"));
		style = gtk_widget_get_style(textview->text);
		quote_color = uri_color = emphasis_color = style->black;
	}

	textview_set_font();
}

void textview_show_message(TextView *textview, MimeInfo *mimeinfo,
			   const gchar *file)
{
	FILE *fp;

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return;
	}

	textview_show_part(textview, mimeinfo, fp);

	fclose(fp);
}

void textview_show_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
{
	GtkText *text = GTK_TEXT(textview->text);
	gchar buf[BUFFSIZE];
	gchar *boundary = NULL;
	gint boundary_len = 0;
	gboolean charset_is_jis = TRUE;

	g_return_if_fail(fp != NULL);

	textview_clear(textview);

	if (mimeinfo) {
		if (mimeinfo->mime_type == MIME_MULTIPART) {
			if (mimeinfo->sub != NULL) {
				mimeinfo = mimeinfo->sub;
				if (fseek(fp, mimeinfo->fpos, SEEK_SET) < 0) {
					perror("fseek");
					return;
				}
			} else
				return;
		}
		if (mimeinfo->parent && mimeinfo->parent->boundary) {
			boundary = mimeinfo->parent->boundary;
			boundary_len = strlen(boundary);
		}
	}

	gtk_text_freeze(text);

	if (!boundary)
		textview_scan_header(textview, fp);
	else
		while (fgets(buf, sizeof(buf), fp) != NULL)
			if (buf[0] == '\r' || buf[0] == '\n') break;

	if (mimeinfo && mimeinfo->mime_type == MIME_MESSAGE_RFC822) {
		textview_scan_header(textview, fp);
		mimeinfo = mimeinfo->sub;
	}

	if (mimeinfo && mimeinfo->charset) {
		if (!strcasecmp(mimeinfo->charset, "ISO-2022-JP") ||
		    !strcasecmp(mimeinfo->charset, "ISO-2022-JP-2"))
			charset_is_jis = TRUE;
		else
			charset_is_jis = FALSE;
	}

	if (mimeinfo &&
	    (mimeinfo->encoding_type == ENC_QUOTED_PRINTABLE ||
	     mimeinfo->encoding_type == ENC_BASE64)) {
		FILE *tmpfp;

		tmpfp = procmime_decode_content(NULL, fp, mimeinfo);
		if (tmpfp) {
			while (fgets(buf, sizeof(buf), tmpfp) != NULL)
				textview_write_line(textview, buf,
						    charset_is_jis);
			fclose(tmpfp);
		}
	} else {
		if (!boundary) {
			while (fgets(buf, sizeof(buf), fp) != NULL)
				textview_write_line
					(textview, buf, charset_is_jis);
		} else {
			while (fgets(buf, sizeof(buf), fp) != NULL &&
			       !IS_BOUNDARY(buf, boundary, boundary_len))
				textview_write_line
					(textview, buf, charset_is_jis);
		}
	}

	gtk_text_thaw(text);
}

void textview_show_mime_part(TextView *textview, MimeInfo *partinfo)
{
	GtkText *text = GTK_TEXT(textview->text);

	if (!partinfo) return;

	textview_clear(textview);

	gtk_text_freeze(text);

	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("To save this part, pop up the context menu with\n"), -1);
	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("right click and select `Save as...', or press `y' key.\n\n"), -1);

	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("To display this part as a text message, select\n"), -1);
	gtk_text_insert
		(text, msgfont, NULL, NULL,
		 _("`Display as text', or press `t' key."), -1);

	gtk_text_thaw(text);
}

static void textview_write_line(TextView *textview, const gchar *str,
				gboolean charset_is_jis)
{
	GtkText *text = GTK_TEXT(textview->text);
	gchar buf[BUFFSIZE];
	gchar *p;
	GdkColor *fg_color;

	if (charset_is_jis) {
		//conv_anytoeuc(buf, sizeof(buf), str);
		conv_jistoeuc(buf, sizeof(buf), str);
		conv_unreadable_eucjp(buf);
	} else {
		strcpy(buf, str);
		conv_unreadable_8bit(buf);
	}
	if (prefs_common.conv_mb_alnum) conv_mb_alnum(buf);
	fg_color = NULL;

	/* change color of quotation */
	/* >, foo>, _> ... ok, <foo>, foo bar>, foo-> ... ng */
	if (prefs_common.enable_color &&
	    (p = strchr(buf, '>')) != NULL) {
		if (p == buf || (p - buf < 25 &&
		    memchr(buf, ' ' , p - buf) == NULL &&
		    memchr(buf, '\t', p - buf) == NULL &&
		    memchr(buf, '<' , p - buf) == NULL &&
		    *(p - 1) != '-'))
			fg_color = &quote_color;
		else {
			gchar *bp = buf;

			for (; bp != p; bp++)
				if (!isspace(*bp)) break;
			if (bp == p)
				fg_color = &quote_color;
		}
	}

	/* parse URI */
	if (prefs_common.enable_color &&
	    ((p = strcasestr(buf, "http://"))  != NULL ||
	     (p = strcasestr(buf, "https://")) != NULL ||
	     (p = strcasestr(buf, "ftp://"))   != NULL)) {
		gchar *ep;
		RemoteURI *uri;

		uri = g_new(RemoteURI, 1);

		gtk_text_insert(text, msgfont, fg_color, NULL, buf, p - buf);

		/* find end point of URI */
		for (ep = p; *ep; ep++)
			if (!isgraph(*ep) || !isascii(*ep) ||
			    strchr("()<>\"", *ep))
				break;

		uri->uri = g_strndup(p, ep - p);
		uri->start = gtk_text_get_point(text);
		gtk_text_insert(text, msgfont, &uri_color, NULL, p, ep - p);
		uri->end = gtk_text_get_point(text);

		textview->uri_list = g_slist_append(textview->uri_list, uri);

		gtk_text_insert(text, msgfont, fg_color, NULL, ep, -1);
	} else
		gtk_text_insert(text, msgfont, fg_color, NULL, buf, -1);
}

void textview_clear(TextView *textview)
{
	GtkText *text = GTK_TEXT(textview->text);

	gtk_text_freeze(text);
	gtk_text_backward_delete(text, gtk_text_get_length(text));
	gtk_text_thaw(text);

	textview_uri_list_remove_all(textview->uri_list);
	textview->uri_list = NULL;
}

void textview_quote_color_set(void)
{
}

void textview_set_font(void)
{
	if (prefs_common.messagefont)
		msgfont = gdk_fontset_load(prefs_common.messagefont);
	//if (!normalfont)
	//	normalfont = gdk_fontset_load(NORMAL_FONT);
	if (!boldfont)
		boldfont = gdk_fontset_load(BOLD_FONT);
}

enum
{
	H_DATE		= 0,
	H_FROM		= 1,
	H_TO		= 2,
	H_NEWSGROUPS	= 3,
	H_SUBJECT	= 4,
	H_CC		= 5,
	H_REPLY_TO	= 6,
	H_FOLLOWUP_TO	= 7,
	H_X_MAILER	= 8,
	H_X_NEWSREADER	= 9,
	H_USER_AGENT	= 10,
	H_ORGANIZATION	= 11,
};

static void textview_scan_header(TextView *textview, FILE *fp)
{
	static HeaderEntry hentry[] = {{"Date: ",         NULL, FALSE},
				       {"From: ",         NULL, TRUE},
				       {"To: ",	          NULL, FALSE},
				       {"Newsgroups: ",   NULL, FALSE},
				       {"Subject: ",      NULL, TRUE},
				       {"Cc: ",	          NULL, FALSE},
				       {"Reply-To: ",     NULL, FALSE},
				       {"Followup-To: ",  NULL, FALSE},
				       {"X-Mailer: ",     NULL, TRUE},
				       {"X-Newsreader: ", NULL, TRUE},
				       {"User-Agent: ",   NULL, TRUE},
				       {"Organization: ", NULL, TRUE},
				       {NULL,	          NULL, FALSE}};
	gchar buf[BUFFSIZE], tmp[BUFFSIZE];
	GtkText *text = GTK_TEXT(textview->text);
	gint hnum;
	HeaderEntry *hp;
	MsgInfo msginfo;

	g_return_if_fail(fp != NULL);

	memset(&msginfo, 0, sizeof(MsgInfo));

	/* show specified headers */
	while ((hnum = procheader_get_one_field(buf, sizeof(buf), fp, hentry))
	       != -1) {
		hp = hentry + hnum;

		gtk_text_insert(text, boldfont, NULL, NULL, buf,
				strlen(hp->name));
		UnMimeHeader(buf);
		conv_jistoeuc(tmp, sizeof(tmp), buf + strlen(hp->name));
		conv_unreadable_eucjp(tmp);
		if (hnum == H_X_MAILER && prefs_common.enable_color &&
		    strcasestr(tmp, "Sylpheed") != NULL)
			gtk_text_insert(text, normalfont, &emphasis_color,
					NULL, tmp, -1);
		else
			gtk_text_insert(text, normalfont, NULL, NULL, tmp, -1);
		gtk_text_insert(text, NULL, NULL, NULL, "\n", 1);
	}

	gtk_text_insert(text, NULL, NULL, NULL, "\n", 1);
}

static void textview_key_pressed(GtkWidget *widget, GdkEventKey *event,
				 TextView *textview)
{
	SummaryView *summaryview =
		textview->messageview->mainwin->summaryview;

	if (!event) return;

	switch (event->keyval) {
	case GDK_Tab:
	case GDK_Home:
	case GDK_Left:
	case GDK_Up:
	case GDK_Right:
	case GDK_Down:
	case GDK_Page_Up:
	case GDK_Page_Down:
	case GDK_End:
	case GDK_Control_L:
	case GDK_Control_R:
		break;
	default:
		summary_pass_key_press_event(summaryview, event);
		gtk_widget_grab_focus(summaryview->ctree);
		break;
	}
}

static void textview_button_pressed(GtkWidget *widget, GdkEventButton *event,
				    TextView *textview)
{
	if (event &&
	    ((event->button == 1 && event->type == GDK_2BUTTON_PRESS) ||
	     event->button == 2)) {
		GSList *cur;
		guint current_pos = GTK_EDITABLE(textview->text)->current_pos;

		for (cur = textview->uri_list; cur != NULL; cur = cur->next) {
			RemoteURI *uri = (RemoteURI *)cur->data;

			if (current_pos >= uri->start &&
			    current_pos < uri->end)
				textview_open_uri(uri->uri);
		}
	}
}

static void textview_uri_list_remove_all(GSList *uri_list)
{
	while (uri_list != NULL) {
		if (uri_list->data) {
			g_free(((RemoteURI *)uri_list->data)->uri);
			g_free(uri_list->data);
		}
		uri_list = g_slist_remove(uri_list, uri_list->data);
	}
}

static void textview_open_uri(const gchar *uri)
{
	static gchar *default_cmdline = "netscape -remote 'openURL(%s,raise)'";
	gchar buf[1024];
	gchar *p;

	if (prefs_common.uri_cmd &&
	    (p = strchr(prefs_common.uri_cmd, '%')) && *(p + 1) == 's' &&
	    !strchr(p + 2, '%'))
		g_snprintf(buf, sizeof(buf) - 1, prefs_common.uri_cmd, uri);
	else {
		if (prefs_common.uri_cmd)
			g_warning(_("Open URI command line is invalid: `%s'"),
				  prefs_common.uri_cmd);
		g_snprintf(buf, sizeof(buf) - 1, default_cmdline, uri);
	}

	if (buf[strlen(buf) - 1] != '&') strcat(buf, "&");
	system(buf);
}
